require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/os/activity_action/activity_action_attributes'
require 'uu/os/activity_action/activity_action_create'
require 'uu/os/activity_action/activity_action_get_activity_action_list'


# Module Unicorn Universe.
module UU
  # Module Operating System.
  module OS
    # Module Activity Action.
    module ActivityAction
      extend self

      # Service path
      PATH = 'uu/os/ActivityAction'

      # Creates a new activity action for the specified entity (activity template or activity instance).
      # 
      # @param [String, UU::OS::UESURI] entity_uri <b>Not nil.</b> URI of the activity template or activity instance for which the new activity action is to be created.
      # @param [UU::OS::ActivityAction::ActivityActionCreate] options Specific options.
      # @option options [String] :code Code of the new activity action. If it is not specified, the code is generated automatically using the system sequence. The code must be unique in the scope of the given meta artifact, otherwise the create command fails.
      # @option options [String, UU::OS::ActivityAction::ActivityActionType] :type <b>Not nil.</b> Type of the new activity action. All possible types can be found as constants in {UU::OS::ActivityAction::ActivityActionType} class.
      # @option options [String, UU::OS::UESURI] :activity_state_template_uri <b>Not nil.</b> URI of an activity state template which is related to the activity template or its instance set as the main entity. When the main entity is set into the state represented by this state template, the created activity action will be executed. It is not allowed to use a state template that is already used by another action for setting artifact state.
      # @option options [String, UU::OS::UESURI] :target_artifact_state_template_uri <b>Not nil if the type is SET_ARTIFACT_STATE_ACTION.</b> URI of an artifact state template which is related to the specific meta artifact according to the entered main entity. It represents a state that will be set to an artifact created according to the related meta artifact after this activity action is executed. This parameter must not represent a system state.
      # @option options [String, UU::OS::UESURI] :target_activity_state_template_uri <b>Not nil if the type is SET_ACTIVITY_STATE_ACTION.</b> URI of an activity state template. It represents a state that will be set to the given activity after this activity action is executed. It must not belong to a generic activity template unless the main entity is also generic.
      # @option options [String, UU::OS::UESURI] :macro_uri <b>Not nil if the type is EXEC_MACRO_ACTION.</b> URI of an artifact created according to a meta artifact for macro. This artifact represents a macro that will be run after this activity action is executed. This artifact must be in an active state.
      #
      # @return [UU::OS::UESURI] URI of the new activity action.
      #
      # @example
      #   # Create an activity action for setting an artifact state
      #   UU::OS::ActivityAction.create('ues:TERRITORY:ARTIFACT:ENTITY',
      #      :code => 'Activity action code',
      #      :type => 'SET_ARTIFACT_STATE_ACTION',
      #      :activity_state_template_uri => 'ues:TERRITORY:META_ARTIFACT:ACTIVITY_STATE_TEMPLATE',
      #      :target_artifact_state_template_uri => 'ues:TERRITORY:META_ARTIFACT:ARTIFACT_STATE_TEMPLATE')
      #
      #   # Create an activity action for setting an activity state
      #   UU::OS::ActivityAction.create('ues:TERRITORY:ARTIFACT:ENTITY',
      #      :code => 'Activity action code',
      #      :type => 'SET_ACTIVITY_STATE_ACTION',
      #      :activity_state_template_uri => 'ues:TERRITORY:META_ARTIFACT:ACTIVITY_STATE_TEMPLATE',
      #      :target_activity_state_template_uri => 'ues:TERRITORY:META_ARTIFACT:ACTIVITY_STATE_TEMPLATE')
      #
      #   # Create an activity action for executing a macro
      #   UU::OS::ActivityAction.create('ues:TERRITORY:ARTIFACT:ENTITY',
      #      :code => 'Activity action code',
      #      :type => 'EXEC_MACRO_ACTION',
      #      :activity_state_template_uri => 'ues:TERRITORY:META_ARTIFACT:ACTIVITY_STATE_TEMPLATE',
      #      :macro_uri => 'ues:TERRITORY:MACRO')
      def create(entity_uri, options)
        svc = UU::OS::REST::RemoteClient.new(ActivityAction)
        payload = UU::OS::ActivityAction::ActivityActionCreate.new(options).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('create', entity_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Returns attributes of the specified activity action.
      #
      # @param [String, UU::OS::UESURI] activity_action_uri <b>Not nil.</b> URI of the activity action whose attributes are to be returned.
      #
      # @return [UU::OS::ActivityAction::ActivityActionAttributes] Object with attributes of the activity action.
      #
      # @example
      #   # Get attributes of an activity action
      #   UU::OS::ActivityAction.get_attributes('ues:TERRITORY:ARTIFACT:ACTIVITY_ACTION')
      def get_attributes(activity_action_uri)
        svc = UU::OS::REST::RemoteClient.new(ActivityAction)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getAttributes', activity_action_uri)
          return ActivityAction::ActivityActionAttributes.new(res)
        end
      end

      # Deletes the specified activity action.
      #
      # @param [String, UU::OS::UESURI] activity_action_uri <b>Not nil.</b> URI of the activity action which is to be deleted.
      #
      # @example
      #   # Delete an activity action
      #   UU::OS::ActivityAction.delete('ues:TERRITORY:ARTIFACT:ACTIVITY_ACTION')
      def delete(activity_action_uri)
        svc = UU::OS::REST::RemoteClient.new(ActivityAction)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('delete', activity_action_uri)
        end
      end

      # Returns a list of activity actions of the specified entity (activity template or activity instance)
      # according to the entered criteria.
      #
      # @param [String, UU::OS::UESURI] entity_uri <b>Not nil.</b> URI of the entity (activity template or activity instance) whose activity action list is to be returned.
      # @param [UU::OS::ActivityAction::ActivityActionGetActivityActionList] criteria Specific options.
      # @option criteria [String] :query The query for filtering and ordering of the result list. It has to conform the ues query syntax (see documentation for more info). If it is not specified, no filtering is applied and entries are ordered by the parameter type and code as the second criteria. Otherwise, it is possible to filter and sort the result list by the following attributes: type, code.
      #
      # @see https://uu.unicornuniverse.eu/ues/sesm?&SessFree=ues:UNI-BT:UES.DOC%2FCMD_V1%2FUESQUERY UESQuery Documentation
      #
      # @return [UU::OS::REST::ResultList<ActivityAction::ActivityActionGetActivityActionList, ActivityAction::ActivityActionAttributes>] List of activity actions according to the entered entity. Contains instances of {UU::OS::ActivityAction::ActivityActionAttributes}.
      #
      # @example
      #   # Return list of execute macro actions returned in descending order by code
      #   UU::OS::ActivityAction.get_activity_action_list('ues:TERRITORY:ARTIFACT:ENTITY',
      #      :query => 'type="EXEC_MACRO_ACTION" ORDER BY code DESC')
      def get_activity_action_list(entity_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(ActivityAction)
        dto = ActivityAction::ActivityActionGetActivityActionList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getActivityActionList', entity_uri)
          return UU::OS::REST::ResultList.new(ActivityAction::ActivityActionGetActivityActionList, ActivityAction::ActivityActionAttributes, res)
        end
      end

    end
  end

end
