require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/os/sheet/mime_type'
require 'uu/os/sheet/sheet_attributes'
require 'uu/os/sheet/sheet_get_sheet_list'
require 'uu/os/sheet/sheet_check_out'
require 'uu/os/sheet/sheet_check_in'
require 'uu/os/sheet/sheet_get_sheet_data'
require 'uu/os/sheet/sheet_create'
require 'uu/os/sheet/sheet_get_component_content'
require 'uu/os/sheet/sheet_get_component_data'
require 'uu/os/sheet/sheet_replace_component'
require 'uu/os/sheet/sheet_delete_component'
require 'uu/os/sheet/sheet_set_attributes'
require 'uu/os/sheet/sheet_get_version_attributes'
require 'uu/os/sheet/sheet_version_attributes'
require 'uu/os/sheet/sheet_get_version_list'
require 'uu/os/sheet/sheet_delete_version'
require 'uu/os/rest/binary_value'
require 'uu/os/sheet/sheet_get_image_data'
require 'uu/os/sheet/sheet_get_version_data'

module UU
  module OS
    module Sheet
      extend self

      # Service path
      PATH = 'ues/core/content/UESSheet'
      # Returns attributes representing the sheet specified by the <code>sheetUri</code>
      # attribute.
      #
      # @param [String, UU::OS::UESURI] sheet_uri URI of sheet
      #
      # @return [SheetAttributes] DTO withsheet attributes
      def get_attributes(sheet_uri)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getAttributes', sheet_uri)
          return Sheet::SheetAttributes.new(res)
        end
      end

      # This command returns a list of sheets on the specified artifact. Command does not
      # return all sheets, only a sublist. The list of returned sheets couldbe filtered
      # by the name, code, main, order of a sheet and state of the lock. The list is sorted
      # by the name (and codes in case are equal) of asheet by default, or it can be sorted
      # arbitrarily by any of sheet attributes usable for the filtering or these attributes
      # combinations.
      #
      # @param [String, UU::OS::UESURI] artifact_uri URI of artifact.
      # @param [SheetGetSheetList] criteria Criteria
      # for selecting sheets. Available parameters are: {SheetAttributes#name name},
      # {SheetAttributes#code code}, {SheetAttributes#order order},
      # {SheetAttributes#main main}, {SheetAttributes#locked locked},
      # {SheetAttributes#lockTime lockTime}, {SheetAttributes#visible visible} and
      # {SheetAttributes#width width} : {SheetWidthType}.
      #
      # See {https://uu.unicornuniverse.eu/ues/sesm?&SessFree=ues:UNI-BT:UES.DOC%2FCMD_V1%2FUESQUERY UESQuery Documentation}.
      #
      # @return List with sheets of artifact
      def get_sheet_list(artifact_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        dto = Sheet::SheetGetSheetList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getSheetList', artifact_uri)
          return UU::OS::REST::ResultList.new(Sheet::SheetGetSheetList, Sheet::SheetAttributes, res)
        end
      end

      # This command locks the sheet for further modification, locks the given sheet of
      # artifact. When lock succeed or sheet is currently locked by the user,sheet data
      # can be returned. Sheet lock failure throws UESSheetRTException.
      #
      # @param [String, UU::OS::UESURI] sheet_uri URI of sheet to be locked.
      # @param [SheetCheckOut] sheet Specify when stream will be returned and which type of data
      #
      # @return [UU::OS::REST::BinaryValue] DTO contains content of sheet
      def check_out(sheet_uri, sheet = nil)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        dto = UU::OS::Sheet::SheetCheckOut.new(sheet)
        payload = dto.to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          if dto.lock_only == TRUE || dto.return_stream_handler == FALSE
            svc.post('checkOut', sheet_uri, payload)
            return nil
          else
            return svc.post_binary('checkOut', sheet_uri, payload)
          end
        end
      end

      # This command changes the content of the public sheet version and unlocks it by default.
      # Sheet content is in UXML or UDF format, data are streamed.Sheet must be unlocked
      # or the user must be the lock owner.
      #
      # If using XML format document cannot contain links to property of object from other territory.
      #
      # @param [String, UU::OS::UESURI] sheet_uri URI of the sheet.
      # @param [SheetCheckIn] sheet Attributes of importing sheet
      def check_in(sheet_uri, sheet = nil)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        dto = UU::OS::Sheet::SheetCheckIn.new(sheet)
        payload = dto.to_hash
        payload.delete(:content)
        tmpfile = Tempfile.new("SheetCheckIn")
        tmpfile.binmode
        begin
          if (!dto.content.kind_of?(UU::OS::REST::BinaryValue))
            dto.content = UU::OS::REST::BinaryValue.new(dto.content)
          end
          if (dto.content.data.respond_to?(:read))
            if (dto.content.data.respond_to?(:path))
              payload[:content] = dto.content.to_hash(false)
            else
              dto.content.data.rewind
              tmpfile.write(dto.content.data.read)
              payload[:content] = tmpfile
            end
          else
            tmpfile.write(dto.content.data)
            payload[:content] = tmpfile
          end
          tmpfile.flush
          tmpfile.rewind

          UU::OS::QoS::QoSHandler.auto_retry do
            svc.post('checkIn', sheet_uri, payload)
          end
        ensure
        tmpfile.close
        tmpfile.unlink
        end
      end

      # Returns the content of the sheet specified by <code>sheetUri</code>. Data are streamed.
      # Available output data formats are UXML file, UDF archive, and a zip archive
      # with the sheet content exported as HTML document
      # (see {SheetGetSheetData} for more details).
      #
      # @param [String, UU::OS::UESURI] sheet_uri URI of the sheet.
      # @param [SheetGetSheetData] sheet Specify when stream will be returned and which type of data
      #
      # @return [UU::OS::REST::BinaryValue] DTO with sheet content.
      #
      # @example
      #   # Get sheet content as UXML (uri needs to be replaced with one of an existing sheet for the example to work)
      #   UU::OS::Sheet.get_sheet_data('ues:TERRITORY:ARTIFACT:SHEET',
      #   :requested_mime_type=>UU::OS::Sheet::MimeType::UXML)
      def get_sheet_data(sheet_uri, sheet = nil)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        dto = UU::OS::Sheet::SheetGetSheetData.new(sheet)
        svc.add_parameter('requestedMimeType', dto.requested_mime_type)
        UU::OS::QoS::QoSHandler.auto_retry do
          result = svc.get_binary('getSheetData', sheet_uri)
          if (dto.requested_mime_type == UU::OS::Sheet::MimeType::UXML)
            return UU::OS::REST::BinaryValue.new(StringIO.new(result.data.read.force_encoding("UTF-8")),false)
          else
          return result
          end
        end
      end

      # Command creates a new sheet in the specified artifact. Initial sheet content could
      # be specified. Content is streamed.
      #
      # @param [String, UU::OS::UESURI] artifact_uri URI of artifact
      # @param [SheetCreate] sheet DTO with sheet content
      #
      # @return [UU::OS::UESURI] UESURI of created sheet
      def create(artifact_uri, sheet = nil)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        dto = UU::OS::Sheet::SheetCreate.new(sheet)
        payload = dto.to_hash
        payload.delete(:content)
        tmpfile = Tempfile.new("SheetCreate")
        tmpfile.binmode
        begin
          if (!dto.content.nil?)
            if (!dto.content.kind_of?(UU::OS::REST::BinaryValue))
              dto.content = UU::OS::REST::BinaryValue.new(dto.content)
            end
            if (dto.content.data.respond_to?(:read))
              if (dto.content.data.respond_to?(:path))
                payload[:content] = dto.content.to_hash(false)
              else
                dto.content.data.rewind
                tmpfile.write(dto.content.data.read)
                payload[:content] = tmpfile
              end
            else
              tmpfile.write(dto.content.data)
              payload[:content] = tmpfile
            end
          end
          tmpfile.flush
          tmpfile.rewind

          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.post('create', artifact_uri, payload)
            return UU::OS::UESURI.new(res)
          end
        ensure
        tmpfile.close
        tmpfile.unlink
        end
      end

      # TODO priority 3
      #      # Returns the content of specified sheet component in the BB code format. The result
      #      # is streamed.
      #      #
      #      # @param [String, UU::OS::UESURI] sheet_uri UESURI of the sheet
      #      # @param [SheetGetComponentContent] sheet component id
      #      #
      #      # @return [] DTO containing the component content in a stream
      #      def get_component_content(sheet_uri, sheet = nil)
      #        svc = UU::OS::REST::RemoteClient.new(Sheet)
      #        payload = UU::OS::Sheet::SheetGetComponentContent.new(sheet).to_json
      #
      #        UU::OS::QoS::QoSHandler.auto_retry do
      #          svc.get('getComponentContent', sheet_uri, payload)
      #        end
      #      end

      # Returns the content of specified sheet component. The result
      # is streamed in the BB code format by default. Available output data formats are UXML or BB code.
      #
      # @param [String, UU::OS::UESURI] sheet_uri UESURI of the sheet
      # @param [SheetGetComponentData] sheet Specify which component data will be returned and type of data
      #
      # @return [UU::OS::REST::BinaryValue] DTO containing the component content in a stream
      def get_component_data(sheet_uri, sheet = nil)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        dto = UU::OS::Sheet::SheetGetComponentData.new(sheet)
        svc.add_parameter("componentId", dto.component_id)
        svc.add_parameter("requestedMimeType", dto.requested_mime_type)

        UU::OS::QoS::QoSHandler.auto_retry do
          return svc.get_binary('getComponentData', sheet_uri)
        end
      end

      # The uuCommand replaces specified component with given content.
      #
      # @param [String, UU::OS::UESURI] sheet_uri UESURI of the sheet.
      # @param [UU::OS::Sheet::SheetReplaceComponent] sheet Object specifying component ID, new content
      #   and MIME type
      # @option sheet [String] :component_id Identification code of the component in the given sheet
      #   (id of a paragraph etc.). Cannot be null.
      # @option sheet [UU::OS::REST::BinaryValue] :content New content which is to be inserted at
      #   the component position. It must be in BB code format and cannot be null.
      # @option sheet [String] :requested_mime_type Specifies the format of data in the input stream.
      #   Only BBCODE is supported at the moment. If null is set, MIME type must not be null on content attribute.
      # @option sheet [TrueClass, FalseClass] :preserve_lock_state When true is set,
      #   the command preserves the state of the sheet lock. Therefore if the sheet is locked before
      #   the command execution, it also remains locked after the execution. When false or null is set,
      #   the sheet is unlocked after the command execution.
      #
      # @return [Array] List of IDs of the inserted components.
      #
      # @example
      #   # Replace the content of a paragraph (uri needs to be replaced with one of an existing sheet for the example to work)
      #   UU::OS::Sheet.replace_component('ues:TERRITORY:ARTIFACT:SHEET',
      #   :component_id=>"583c10bfdbd326ba1ebfbd146d29e6392-7fff",
      #   :content=>REST::BinaryValue.new(:data=>'[ues.p]New content.[/ues.p]'),
      #   :requested_mime_type=>'BBCODE')
      def replace_component(sheet_uri, sheet = nil)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        payload = UU::OS::Sheet::SheetReplaceComponent.new(sheet).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('replaceComponent',sheet_uri, payload)
          return JSON.parse(res).values[0]
        end
      end

      # Publishes private version and opens the lock sheet. All changes made in private
      # version are published.
      #
      # @param [String, UU::OS::UESURI] sheet_uri UESURI of the sheet.
      def publish(sheet_uri)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('publish', sheet_uri)
        end
      end

      # Unlocks the specified sheet. Allows to unlock the sheet locked by another user
      #
      # @param [String, UU::OS::UESURI] sheet_uri UESURI of the sheet.
      def unlock(sheet_uri)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('unlock', sheet_uri)
        end
      end

      # Deletes the sheet from the system.
      #
      # @param [String, UU::OS::UESURI] sheet_uri UESURI of the sheet which is about to be deleted.
      def delete(sheet_uri)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('delete', sheet_uri)
        end
      end

      # Deletes specified component from the sheet.
      #
      # @param [String, UU::OS::UESURI] sheet_uri UESURI of the sheet.
      # @param [SheetDeleteComponent] sheet DTO containing attributes for deletingcomponent.
      def delete_component(sheet_uri, sheet = nil)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        payload = UU::OS::Sheet::SheetDeleteComponent.new(sheet).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('deleteComponent', sheet_uri, payload)
        end
      end

      # This command updates basic attributes of an sheet.
      #
      # @param [String, UU::OS::UESURI] sheet_uri UESURI of the sheet whose attributes are to be changed.
      # @param [SheetSetAttributes] sheet DTO containingthe sheet attributes.
      #
      # @return [UU::OS::UESURI] UESURI of the changed sheet.
      def set_attributes(sheet_uri, sheet = nil)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        payload = UU::OS::Sheet::SheetSetAttributes.new(sheet).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('setAttributes', sheet_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # The uuCommand returns attributes of given sheet version.
      #
      # @param [String, UU::OS::UESURI] sheet_uri UESURI of the sheet whose version is being searched.
      # @param [SheetGetVersionAttributes] version Object containing ID of the searched version.
      # @option version [Integer] :version_id ID of the searched sheet version. It cannot be left null or blank.
      #
      # @return [SheetVersionAttributes] Object with attributes of the searched version.
      #
      # @example
      #   # Get sheet version attributes (uri needs to be replaced with one of an existing sheet for the example to work)
      #   attributes = UU::OS::Sheet.get_version_attributes('ues:TERRITORY:ARTIFACT:SHEET', :version_id => 123456789)
      def get_version_attributes(sheet_uri, version = nil)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        dto = UU::OS::Sheet::SheetGetVersionAttributes.new(version)
        svc.add_parameter('versionID', dto.version_id)

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getVersionAttributes', sheet_uri)
          return Sheet::SheetVersionAttributes.new(res)
        end
      end

      # The uuCommand returns a list of sheet versions of the specified sheet.
      #
      # @param [String, UU::OS::UESURI] sheet_uri URI of the sheet whose versions will be retrieved.
      # @param [SheetGetVersionList] criteria Criteria for selecting sheets.
      # @option criteria [String] :query Query for filtering and ordering of the result list (more information in {uuQuery - Guideline}[https://plus4u.net/ues/sesm?SessFree=ues%253AUNI-BT%255B210795%255D%253AUES.DOC%252FCMD_V1%252FUESQUERY%255B44191574519623267%255D%253A]). It is possible to filter the result list by label, creationTime, creationStamp and revision.
      #
      # @return [Array] An array with the command result. Contains instances of {UU::OS::Sheet::SheetVersionAttributes}.
      #
      # @example
      #   # Get sheet versions ordered by revision number in ascending order (uri needs to be replaced with one of an existing sheet for the example to work)
      #   list = UU::OS::Sheet.get_version_list('ues:TERRITORY:ARTIFACT:SHEET', :query => "ORDER BY revision ASC")
      def get_version_list(sheet_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        dto = Sheet::SheetGetVersionList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getVersionList', sheet_uri)
          return UU::OS::REST::ResultList.new(Sheet::SheetGetVersionList, Sheet::SheetVersionAttributes, res)
        end
      end

      # Deletes the version from the system.
      #
      # @param [String, UU::OS::UESURI] sheet_uri UESURI of the sheet whose version is about to be deleted.
      # @param [SheetDeleteVersion] version Object containing ID of the version which is to be deleted.
      # @option version [Integer] :version_id ID of the sheet version. It cannot be left nil or blank.
      #
      # @example
      #   # Delete sheet version (uri needs to be replaced with one of an existing sheet for the example to work)
      #   UU::OS::Sheet.delete_version('ues:TERRITORY:ARTIFACT:SHEET', :version_id => 123456789)
      def delete_version(sheet_uri, version = nil)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        payload = UU::OS::Sheet::SheetDeleteVersion.new(version).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('deleteVersion', sheet_uri, payload)
        end
      end

      # Returns streamed data of the image from the specified content.
      #
      # @param [String, UU::OS::UESURI] sheet_uri <b>Not nil.</b> UESURI of the sheet containing the image which should be returned.
      # @param [SheetGetImageData] image Specific options.
      # @option image [String] :image_id <b>Not nil.</b> ID of the image from the specified sheet. It can be found in the attribute UUID of the element core:image contained in XML (UUXML) representation of the sheet.
      #
      # @example
      #   # Get image identified by UUID a89ad302b98373e4140188114f2756a020-7fff from the sheet (main entity).
      #   UU::OS::Sheet.get_image_data('ues:TERRITORY:ARTIFACT:SHEET', :image_id => 'a89ad302b98373e4140188114f2756a020-7fff')
      def get_image_data(sheet_uri, image)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        dto = UU::OS::Sheet::SheetGetImageData.new(image)
        svc.add_parameter('image', dto.image_id)

        UU::OS::QoS::QoSHandler.auto_retry do
          return svc.get_binary('getImageData',sheet_uri)
        end
      end

      # Retrieves the specified historical version of the given sheet in xml format.
      #
      # @param [String, UU::OS::UESURI] sheet_uri <b>Not nil.</b> UESURI of the sheet, which content will be retrieved.
      # @param [SheetGetVersionData] version Specific options.
      # @option version [String] :version_id <b>Not nil.</b> Id of version which data will be retrieved by command.
      # @option version [String] :requested_mime_type Format of data retrieved by command. Only possible value is 'application/ues-content-xml'.
      #
      # @return [UU::OS::REST::BinaryValue] Binary data containing versions content depending on requested mime type.
      #
      # @example
      #   # Get version of the sheet 'ues:TER:ART:SHEET' identified by its OID = 456232122
      #   UU::OS::Sheet.get_version_data('ues:TER:ART:SHEET', :versionID => "456232122", :requestedMimeType => "application/ues-content-xml")
      def get_version_data(sheet_uri, version)
        svc = UU::OS::REST::RemoteClient.new(Sheet)
        dto = UU::OS::Sheet::SheetGetVersionData.new(version)
        svc.add_parameter('requestedMimeType', dto.requested_mime_type)
        svc.add_parameter('versionID', dto.version_id)
        UU::OS::QoS::QoSHandler.auto_retry do
          result = svc.get_binary('getVersionData', sheet_uri)
          if (dto.requested_mime_type == UU::OS::Sheet::MimeType::UXML)
            return UU::OS::REST::BinaryValue.new(StringIO.new(result.data.read.force_encoding("UTF-8")),false)
          else
          return result
          end
        end
      end

    end
  end
end
